// SPDX-License-Identifier: Apache-2.0

#pragma once

#include <string_view>

constexpr std::string_view GIT_COMMIT        = "";
constexpr std::string_view COMPILER          = "g++";
constexpr std::string_view COMPILER_VERSION  = "15.2.1";
constexpr std::string_view CONFIGURE_COMMAND = "make STATIC= GPU_SUPPORT=true RSMI_STATIC=";
